/*
Copyright 1989 by the Movies of the Future Group
of the MIT Media Laboratory
All rights reserved.

Permission to use, copy, and modify this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of M.I.T. not be used in
advertising or publicity pertaining to distribution of the software
without specific, written prior permission.  This software or any
other copies thereof may not be provided or otherwise made available to
others.  No title to or ownership of the software is hereby transferred.
M.I.T. makes no representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.
*/
/* EMACS_MODES: c */

/* Convert between any of data types buUcdDfFzZxX.

	May be used independent of datfiles.

	The following rules are followed:

	Real->complex	puts 0 in imaginary part
	Complex->real	retains only real part
	Float->integer	rounds to nearest integer
	Any->integer	clips if out of range

	The fact that ibuf is copied to obuf in order of increasing addresses
	should be observed if these buffers overlap in memory.
*/

#include "dat.h"
/* #include <ip/ip.h> */

/*
** Rewritten by Pat Romano on 3/23/89
** support added for fixed point data conversion
**
** 1/5/94  sbeck  added byte swapped version 
*/

#define CBUFSIZE 		(256)
#define ROUNDUP(x) 		(x + 0.5)
#define ROUNDDN(x) 		(x - 0.5)

/* note: some architectures (like sun) do better this way. -Shawn */
#define MINCHAR			(signed char)(0x80)
#define MAXCHAR			(signed char)(0x7f)
#define MINBYTE			(unsigned char)(0x00)
#define MAXBYTE			(unsigned char)(0xff)
#define MINSHORT		(signed short)(0x8000)
#define MAXSHORT		(signed short)(0x7fff)
#define MINUSHORT		(unsigned short)(0x0000)
#define MAXUSHORT		(unsigned short)(0xffff)
#define MINLONG			(int32)(0x80000000)
#define MAXLONG			(int32)(0x7fffffff)
#define MINULONG		(unsigned int32)(0x00000000)
#define MAXULONG		(unsigned int32)(0xffffffff)

/* same values in straight decimal - sun's can't handle it
#define MINCHAR			(-128)
#define MAXCHAR			(127)
#define MINBYTE			(0)
#define MAXBYTE			(255)
#define MINSHORT		(-32768)
#define MAXSHORT		(32767)
#define MINUSHORT		(0)
#define MAXUSHORT		(65535)
#define MINLONG			(-2147483648)
#define MAXLONG			(2147483647)
#define MINULONG		(0)
#define MAXULONG		(4294967295)
*/

static long clipwarned = FALSE;

void convert_datatype(obuf, otype, ibuf, itype, n)
char 		obuf[]; 
char		ibuf[];
int 		otype; 
int		itype;
register int 	n;
{
    if (n <= 0) 
    {
	return;
    } 

    /*
    ** =============
    ** Straight Copy
    ** =============
    */
    if (itype == otype) 
    {
	if (obuf != ibuf) 
	{
	    register char *src = ibuf; 
	    register char *dst = obuf;

	    n <<= _dsize(itype);

	    while(n--)
	    {
		*dst++ = *src++;
	    } 
	}
    }
    /*
    ** =================
    ** Deal with doubles
    ** =================
    */
    else if (otype == 'F') 
    {
	register double *dst = (double *) obuf;

	switch(itype) 
	{
	    case 'a':
	    case 'c':
	    {
		register char *src = (char *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		}

	    } break;
	    case 'd':
	    {
		register short *src = (short *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		}

	    } break;
	    case 'D':
	    {
		register int32 *src = (int32 *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		}

	    } break;
	    case 'b':
	    {
		register unsigned char *src = (unsigned char *) ibuf;

		while(n--)
		{
		    *dst++ = *src++;
		}

	    } break;
	    case 'u':
	    {
		register unsigned short *src = (unsigned short *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		}

	    } break;
	    case 'U':
	    {
		register unsigned int32 *src = (unsigned int32 *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		}

	    } break;
	    case 'f':
	    {
		register float *src = (float *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		}

	    } break;
	    case 'z':
	    {
		register float *src = (float *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		     src++;
		}

	    } break;
	    case 'Z':
	    {
		register double *src = (double *) ibuf;

		while(n--)
		{
		    *dst++ = (double) *src++;
		     src++;
		}

	    } break;
	    case 'X':
	    {
		register int32 *src = (int32 *) ibuf;

		while(n--)
		{
		    *dst++ = ((double) *src++) / ((double)MAXUSHORT + 1.0);
		}

	    } break;
	    case 'x':
	    {
		register short *src = (short *) ibuf;

		while(n--)
		{
		    *dst++ = ((double) *src++) / ((double)MAXBYTE + 1.0);
		}

	    } break;
	    default: 
		_dsize(itype);	/* produce error message */
	}
    }
    else if (itype == 'F') 
    {
	register double *src = (double *) ibuf;
	register double  x;

	switch(otype) 
	{
	    case 'a': 
	    case 'c': 
	    {
		register char *dst = (char *) obuf;

		while(n--)
		{
		    x = *src++;

		    *dst++ = x < 0.0 				? 
			(x < MINCHAR ? MINCHAR : (char)ROUNDDN(x)) 	:
			(x > MAXCHAR ? MAXCHAR : (char)ROUNDUP(x));
		}
	    } break;
	    case 'd': 
	    {
		register short *dst = (short *) obuf;

		while(n--)
		{
		    x = *src++;

		    *dst++ = x < 0.0 				? 
			(x < MINSHORT ? MINSHORT : (short)ROUNDDN(x)) 	:
			(x > MAXSHORT ? MAXSHORT : (short)ROUNDUP(x));
		}
	    } break;
	    case 'D': 
	    {
		register int32 *dst = (int32 *) obuf;

		while(n--)
		{
		    x = *src++;

		    *dst++ = x < 0.0 				? 
		      (x < MINLONG ? MINLONG : (int32)ROUNDDN(x)) 	:
		      (x > MAXLONG ? MAXLONG : (int32)ROUNDUP(x));
		}
	    } break;
	    case 'b': 
	    {
		register unsigned char *dst = (unsigned char *) obuf;

		while(n--)
		{
		    x = *src++;

		    *dst++ = x < 0.0 ? 0 :
		      (x > MAXBYTE ? MAXBYTE : (unsigned char)ROUNDUP(x));
		}
	    } break;
	    case 'u': 
	    {
		register unsigned short *dst = (unsigned short *) obuf;

		while(n--)
		{
		    x = *src++;

		    *dst++ = x < 0.0 ? 0 :
		      (x > MAXUSHORT ? MAXUSHORT : (unsigned short)ROUNDUP(x));
		}
	    } break;
	    case 'U': 
	    {
		register unsigned int32 *dst = (unsigned int32 *) obuf;

		while(n--)
		{
		    x = *src++;

		    *dst++ = x < 0.0 ? 0 :
		      (x > MAXULONG ? MAXULONG : (unsigned int32)ROUNDUP(x));
		}
	    } break;
	    case 'f': 
	    {
		register float *dst = (float *) obuf;

		while(n--)
		{
		    *dst++ = (float) *src++;
		}
	    } break;
	    case 'z': 
	    {
		register float *dst = (float *) obuf;

		while(n--)
		{
		    *dst++ 	= (float) *src++;
		    *dst++ 	= 0.0;
		}
	    } break;
	    case 'Z': 
	    {
		register double *dst = (double *) obuf;

		while(n--)
		{
		    *dst++ 	= (double) *src++;
		    *dst++ 	= 0.0;
		}
	    } break;
	    case 'X':
	    {
		register int32 *dst = (int32 *) obuf;

		while(n--)
		{
		    x = (*src++) * ((double)MAXUSHORT + 1.0);

		    *dst++ = x < 0.0 			  ? 
		      (x < MINLONG ? MINLONG : (int32)ROUNDDN(x)):
		      (x > MAXLONG ? MAXLONG : (int32)ROUNDUP(x));
		}

	    } break;
	    case 'x':
	    {
		register short *dst = (short *) obuf;

		while(n--)
		{
		    x = (*src++) * ((double)MAXBYTE + 1.0);

		    *dst++ = x < 0.0 			    ? 
		      (x < MINSHORT ? MINSHORT : (short)ROUNDDN(x)):
		      (x > MAXSHORT ? MAXSHORT : (short)ROUNDUP(x));
		}

	    } break;
	    default: 
	        /* to produce an error */
		_dsize(otype);	
	}
    }
    else if (otype == 'Z' && itype == 'z') /* treat complex as 2 reals */
    {
	convert_datatype(obuf, 'F', ibuf, 'f', n << 1);
    }
    else if (otype == 'z' && itype == 'Z')
    {
	convert_datatype(obuf, 'f', ibuf, 'F', n << 1);
    }
    else 
    {			/* use double as intermediate */
	double dbuf[CBUFSIZE];

	int 	icnt; 
	int	ocnt;

	icnt = CBUFSIZE << _dsize(itype);
	ocnt = CBUFSIZE << _dsize(otype);

	while (n > 0) 
	{
	    int cnt = n > CBUFSIZE ? CBUFSIZE : n;

	    convert_datatype((char *)dbuf, 'F', ibuf, itype, cnt);
	    convert_datatype(obuf, otype, (char *)dbuf, 'F', cnt);
	    ibuf 	+= icnt;
	    obuf 	+= ocnt;
	    n 		-= cnt;
	}
    }
}



typedef union {
  unsigned char  b[2];
  unsigned short u;
  signed short   d;
} data2;

typedef union {
  unsigned char  b[4];
  unsigned short u[2];
  signed short   d[2];
  unsigned int32 U;
  signed int32   D;
  float          f;
} data4;

typedef union {
  unsigned char  b[8];
  unsigned short u[4];
  signed short   d[4];
  unsigned int32 U[2];
  signed int32   D[2];
  float          f[2];
  data4         d4[2];
  double         F;
} data8;

/*
#define SWAP_u(tmp) \
  ((unsigned int32)(tmp).b[0]<<8 | (unsigned int32)(tmp).b[1])

#define SWAP_U(tmp) \
  ((unsigned int32)(tmp).b[0]<<24 | (unsigned int32)(tmp).b[1]<<16 | \
   (unsigned int32)(tmp).b[2]<<8 | (unsigned int32)(tmp).b[3])

#define SWAP_F(tmp) \
  { unsigned int32 junk; \
    junk = \
      (unsigned int32)(tmp).b[0]<<24 | (unsigned int32)(tmp).b[1]<<16 | \
      (unsigned int32)(tmp).b[2]<<8 | (unsigned int32)(tmp).b[3]; \
    (tmp).U[0] = \
      (unsigned int32)(tmp).b[4]<<24 | (unsigned int32)(tmp).b[5]<<16 | \
      (unsigned int32)(tmp).b[6]<<8 | (unsigned int32)(tmp).b[7]; \
    (tmp).U[1] = junk; \
  }
*/

/* swap two bytes */
#define SWAP_b(a,b) { register unsigned char _swab; _swab=(a); (a)=(b); (b)=_swab; }

/* swap two int32s */
#define SWAP_U(a,b) { register unsigned int32 _swau; _swau=(a); (a)=(b); (b)=_swau; }


/* swap the 1 pair of bytes within a data2 */
#define SWAB_d2(tmp) { SWAP_b( (tmp).b[0], (tmp).b[1]); }

/* swap the 2 pairs of bytes within a data4 */
#define SWAB_d4(tmp) { SWAP_b( (tmp).b[0], (tmp).b[3]); SWAP_b( (tmp).b[1], (tmp).b[2]); }

/* swap the 2 pairs of bytes within each of the pair of int32's and then swap the pair of int32s in a data8 */
#define SWAB_d8(tmp) { SWAB_d4((tmp).d4[0]); SWAB_d4((tmp).d4[1]); SWAP_U((tmp).U[0],(tmp).U[1]); };


/* this function does inplace ENDIAN conversion on the given buffer
 */
void convert_endian(iobuf, iotype, n)
   char 	iobuf[]; 
   int 		iotype; 
   register int	n;
{
  if (n <= 0) 
    return;

  switch(_dsize(iotype)) 
    {
    case 0: /* a,v,c,b: no byte swapping needed */
      break;
    case 1: /* d,u,x: swap pairs of bytes within each short */
      {
	register unsigned short *p = (unsigned short *)iobuf; 
	register data2 tmp;
	while(n--) { 
	  tmp.u = *p; SWAB_d2(tmp); *p++ = tmp.u;
	}
      }
      break;
    case 2: /* D,U,f,X: swap quads of bytes within each int */
      {
	register unsigned int32 *p = (unsigned int32 *)iobuf; 
	register data4 tmp;
	while(n--) { 
	  tmp.U = *p; SWAB_d4(tmp); *p++ = tmp.U;
	}
      }
      break;
    case 3: 
      { /* z: swap quads's of bytes within each float (int32) */
	if( iotype == 'z' ) {
	  register unsigned int32 *p = (unsigned int32 *)iobuf; 
	  register data4 tmp;
	  while(n--) { 
	    tmp.U = *p; SWAB_d4(tmp); *p++ = tmp.U;
	    tmp.U = *p; SWAB_d4(tmp); *p++ = tmp.U;
	  }
	}
	else { /* F: swap quads's of bytes within each of 2 ints then swap ints */
	  register double *p = (double *)iobuf; 
	  register data8 tmp;
	  while(n--) { 
	    tmp.F = *p; SWAB_d8(tmp); *p++ = tmp.F;
	  }
	}
      }
      break;
    case 4: /* Z: swap quads's of bytes within two sets of two ints */
      {
	register double *p = (double *)iobuf; 
	register data8 tmp;
	while(n--) { 
	  tmp.F = *p; SWAB_d8(tmp); *p++ = tmp.F;
	  tmp.F = *p; SWAB_d8(tmp); *p++ = tmp.F;
	}
      }
      break;
    default:
      exit(_daterrh("convert_endian: dsize:%d not supported\n", _dsize(iotype)));
      /*NOTREACHED*/
      break;
    }
}


/* 
** given a data type, return log2 bytes/sample 
*/
int 
_dsize(dtype)
int dtype;
{
    switch(dtype) 
    {
	case 'a': 
	case 'v': 
	case 'c': 
	case 'b': 
	    return(0);
	case 'd': 
	case 'u': 
	case 'x':
	    return(1);
	case 'D': 
	case 'U':
	case 'f': 
	case 'X':
	    return(2);
	case 'F': 
	case 'z': 
	    return(3);
	case 'Z': 
	    return(4);
	default:
	    exit(_daterrh("_dsize: data type %c. not supported\n", dtype));
		/*NOTREACHED*/
    }
}

